import inspect
import os
from contextlib import contextmanager


def is_degraded():
    return os.environ.get("PYTEST_MPROC_DISABLED")


class _GlobalFixtures:
    count = 0
    finalizers = []
    initializers = []
    session_contexts = []


def global_finalizer(*args):
    """
    Decorator for adding a finalizer called only once and only when all tests complete
    """
    def inner(func):
        _GlobalFixtures.finalizers.append((func, args))
        return func
    return inner


def global_initializer(*args):
    """
    Decorator to add an initializer called once before any tests execute
    """
    def inner(func):
        _GlobalFixtures.initializers.append((func, args))
        return func
    return inner


def global_session_context(*args):

    def wrapper(obj):
        if inspect.isfunction(obj):
            func = contextmanager(obj)
            _GlobalFixtures.session_contexts.append((func, args))
            return func
        elif inspect.isclass(obj) and hasattr(obj, '__enter__') and hasattr(obj, '__exit__'):
            _GlobalFixtures.session_contexts.append((obj, args))
            return obj
        else:
            raise Exception("@pytest_mproc.session_context: object is not a context manager")

    return wrapper



"""
This file contains some standard pytest_* plugin hooks to implement multiprocessing runs
"""
import sys

import contextlib
import inspect
import pytest

from multiprocessing import cpu_count
from pytest_mproc.coordinator import Coordinator
from pytest_mproc.utils import _GlobalFixtures, is_degraded


def parse_numprocesses(s):
    """
    A little bit of processing to get number of parallel processes to use (since "auto" can be used to represent
    # of cores on machine)
    :param s: text to process
    :return: number of parallel worker processes to use
    """
    try:
        if s.startswith("auto"):
            if '*' in s:
                multiplication_factor = int(s.rsplit('*', 1)[-1])
            elif s == "auto":
                multiplication_factor = 1
            else:
                raise Exception("Error: --cores argument must be an integer value or auto or auto*<int factor>")
            return cpu_count() * multiplication_factor
        else:
            return int(s)
    except ValueError:
        raise Exception("Error: --cores argument must be an integer value or \"auto\" or \"auto*<int factor>\"")


@pytest.mark.tryfirst
def pytest_addoption(parser):
    """
    add options to given parser for this plugin
    """
    group = parser.getgroup("pytest_mproc", "better distributed testing through multiprocessing")
    group._addoption(
        "-C",
        "--cores",
        dest="numcores",
        metavar="numcores",
        action="store",
        type=parse_numprocesses,
        help="you can use 'auto' here to set to the number of  CPU cores on host system",
    )
    group._addoption(
            "-D",
            "--disable-mproc",
            dest="mproc_disabled",
            metavar="mproc_disabled",
            action="store",
            type=bool,
            help="disable any parallel mproc testing, overriding all other mproc arguments",
        )


@pytest.mark.tryfirst
def pytest_cmdline_main(config):
    """
    Called before "true" main routine.  This is to set up config values well ahead of time
    for things like pytest-cov that needs to know we are running distributed

    Mostly taken from other implementations (such as xdist)
    """
    if getattr(config.option, "numcores", None) is None or is_degraded() or getattr(config.option, "mproc_disabled"):
        print(">>>>> no number of cores provided or running in environment unsupportive of parallelized testing, "
              "not running multiprocessing <<<<<")
        return
    config.option.numprocesses = config.option.numcores  # this is what pycov uses to determine we are distributed
    # tell xdist not to run, (and BTW setting numprocesses is enough to tell pycov we are distributed)
    config.option.dist = "no"
    val = config.getvalue
    if not val("collectonly"):
        usepdb = config.getoption("usepdb")  # a core option
        if val("dist") != "no":
            if usepdb:
                raise pytest.UsageError(
                    "--pdb is incompatible with distributing tests."
                )  # noqa: E501


@pytest.mark.trylast
def pytest_configure(config):
    if getattr(config.option, "numcores", None) is None or is_degraded() or getattr(config.option, "mproc_disabled"):
        return  # return of None indicates other hook impls will be executed to do the task at hand
    # tell xdist not to run, (and BTW setting numprocesses is enough to tell pycov we are distributed)
    config.option.dist = "no"
    worker = getattr(config.option, "mproc_worker", None)
    if not worker:
        # in main thread,
        # instantiate coordinator here and start to kick off processing on workers early, so they can
        # process config info in parallel to this thread
        config.coordinator = Coordinator(config.option.numcores)
        config.coordinator.start()


@pytest.mark.tryfirst
def pytest_runtestloop(session):
    if session.testsfailed and not session.config.option.continue_on_collection_errors:
        raise session.Interrupted("%d errors during collection" % session.testsfailed)

    if session.config.option.collectonly:
        return  # should never really get here, but for consistency

    worker = getattr(session.config.option, "mproc_worker", None)
    session.config.context_managers = [cm() for cm, args in _GlobalFixtures.session_contexts if not args or
                                       [a for a in args if any([a in item._fixtureinfo.argnames for item in session.items])]] if worker is None else []
    if getattr(session.config.option, "numcores", None) is None or is_degraded() or getattr(session.config.option, "mproc_disabled"):
        for func, args in _GlobalFixtures.initializers:
            try:
                if not args:
                    func()
                else:
                    for item in session.items:
                        if [a for a in args if a in item._fixtureinfo.argnames]:
                            func()
                            break
            except Exception as e:
                print(">>>> ERROR in global initializer. Aborting! <<<<<\n %s" % str(e))
                if hasattr(session.config, 'coordinator'):
                    session.config.coordinator.stop(None)
                raise
        for cm in session.config.context_managers:
            cm.__enter__()
        return

    if worker is None:
        for func, args in _GlobalFixtures.initializers:
            try:
                if not args:
                    func()
                else:
                    for item in session.items:
                        if [a for a in args if a in item._fixtureinfo.argnames]:
                            func()
                            break
            except Exception as e:
                print(">>>> ERROR in global initializer. Aborting! <<<<<\n %s" % str(e))
                if hasattr(session.config, 'coordinator'):
                    session.config.coordinator.stop(None)
                raise
    if getattr(session.config.option, "numcores", None) is None or is_degraded() or getattr(session.config.option, "mproc_disabled"):
        return  # return of None indicates other hook impls will be executed to do the task at hand
    if not session.config.getvalue("collectonly") and worker is None:
        try:
            # main coordinator loop:
            with contextlib.ExitStack() as stack, session.config.coordinator as coordinator:
                for cm in session.config.context_managers:
                        stack.enter_context(cm)
                coordinator.set_items(session.items)
                try:
                    coordinator.run(session)
                except Exception as e:
                    print(">>> ERROR in run loop;  unexpected Exception\n %s" % str(e))
                    raise
        finally:
            session.config.coordinator.stop(None)
    else:
        worker.test_loop(session)
    return True


def pytest_sessionfinish(session, *args):
    worker = getattr(session.config.option, "mproc_worker", None)
    if not worker:
        if getattr(session.config.option, "numcores", None) is None or is_degraded() or getattr(session.config.option,
                                                                                                "mproc_disabled"):
            if hasattr(session.config, "context_managers"):
                for cm in reversed(session.config.context_managers):
                    cm.__exit__(None, None, None)  # exceptions are tracked as part of test reporting, not through cm
        # only executing  in main thread and none of the worker threads:
        for func, args in reversed(_GlobalFixtures.finalizers):
            try:
                if not args:
                    func()
                else:
                    for item in session.items:
                        if [a for a in args if a in item._fixtureinfo.argnames]:
                            func()
                            break
            except Exception:
                print(f">>>> ERROR in global finalizer {func.name};  In=gnoring")



"""
This package contains code to coordinate execution from a main thread to worker threads (processes)
"""

import select
import sys
import time

from _pytest.config import _prepareconfig
from multiprocessing import Process, Queue
from pytest_mproc.worker import main as worker_main


class Coordinator:
    """
    Context manager for kicking of worker Processes to conduct test execution via pytest hooks
    """

    def __init__(self, num_processes):
        """
        :param num_processes: number of parallel executions to be conducted
        """
        self._num_processes = num_processes
        self._tests = []  # set later

        # test_q is for sending test nodeid's to worked
        # result_q is for receiving results as messages, exceptions, test status or any exceptions thrown
        self._result_qs = [Queue() for _ in range(num_processes)]
        self._test_q = Queue()
        self._processes = []
        self._count = 0
        self._session_start_time = time.time()
        self._process_status_text = ["" for _ in range(num_processes)]

    def start(self):
        """
        Start all worker processes

        :return: this object
        """
        for index in range(self._num_processes):
            proc = Process(target=worker_main, args=(index, self._test_q, self._result_qs[index], self._num_processes))
            self._processes.append(proc)
            proc.start()
        return self

    def set_items(self, tests):
        """
        :param tests: the items containing the pytest hooks to the tests to be run
        """
        grouped = [t for t in tests if getattr(t._pyfuncitem.obj, "_pytest_group", None)]
        self._tests = [t for t in tests if t not in grouped]
        groups = {}
        for g in grouped:
            groups.setdefault(g._pyfuncitem.obj._pytest_group, []).append(g)
        for test_list in groups.values():
            self._tests.insert(0, test_list)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def stop(self, delay=None):
        # shouldn't be any procs left, but just in case
        for proc in self._processes:
            if proc:
                try:
                    if delay is not None:
                        proc.join(delay)
                except Exception:
                    pass
                try:
                    proc.terminate()
                except Exception:
                    pass

    @staticmethod
    def _write_sep(s, txt):
        """
        write out text to stdout surrounded by repeated character

        :param s: character to repeat on either side of given text
        :param txt: text to by surrounded
        """
        sep_total = max((70 - 2 - len(txt)), 2)
        sep_len = sep_total // 2
        sep_extra = sep_total % 2
        out = '%s %s %s\n' % (s * sep_len, txt, s * (sep_len + sep_extra))
        sys.stdout.write(out)

    def _output_summary(self):
        """
        Output the summary of test execution
        """
        self._write_sep('=', "STATS")
        for msg in self._process_status_text:
            sys.stdout.write(msg)
        length = sum([1 if not isinstance(t, list) else len(t) for t in self._tests])
        if self._count != length:
            self._write_sep('!', "{} tests unaccounted for {} out of {}".format(length - self._count,
                            self._count, length))
        sys.stdout.flush()

    def _process_worker_message(self, session, typ, data):
        """
        Process a message (as a worker) from the coordinating process

        :param session: the pytest test session
        :param typ: the kind of messsage
        :param data: payload for the message to be processed
        """
        try:
            # TODO: possibly have more structured communications, but for now
            # there are limited types of messages, so sendo message, <data> tuple
            # seems to be fine?
            if typ == 'test_status':
                report = data
                if report.when == 'call' or (report.when == 'setup' and not report.passed):
                    self._count += 1
                session.config.hook.pytest_runtest_logreport(report=report)
                if report.failed:
                    sys.stdout.write("\n%s FAILED\n" % report.nodeid)
            elif typ == 'exit':
                # process is complete, so close it and set to None
                index, worker_count, exitstatus, duration = data  # which process and count of tests run
                name = "worker-%d" % (index + 1)
                self._process_status_text[index] = "Process %s executed %d tests in %.2f seconds\n" % (
                    name, worker_count, duration)
                try:
                    self._processes[index].join(5)
                except Exception as e:
                    pass
                self._processes[index].terminate()
                self._processes[index] = None
                self._result_qs[index].close()
                self._result_qs[index] = None
            elif typ == 'error_message':
                error_msg_text = data
                sys.stdout.write("{}\n".format(error_msg_text))
            elif typ == 'exception':
                # reraise any exception from workers
                raise Exception("Exception in worker process: %s" % str(data))

        except Exception as e:
            sys.stdout.write("INTERNAL_ERROR> %s\n" % str(e))

    def run(self, session):
        """
        Populate test queue and continue to process messages from worker Processes until they complete

        :param session: Pytest test session, to get session or config information
        """
        reader_mapping = {q._reader: q for q in self._result_qs}
        def read_results(timeout=0):
            available_qs = [q._reader for q in self._result_qs if q is not None]
            (inputs, _, _) = select.select(available_qs, [], [], timeout)
            while inputs:
                for input in inputs:
                    items = reader_mapping[input].get()
                    for kind, data in items:
                        self._process_worker_message(session, kind, data)
                available_qs = [q._reader for q in self._result_qs if q is not None]
                (inputs, _, _) = select.select(available_qs, [], [], timeout)

        #while any(self._result_qs):
        for test in self._tests:
            if all([q is None for q in self._result_qs ]):
                raise Exception("Tests remain but no workers left to process them!")
            read_results()
            # Function objects in pytest are not pickle-able, so have to send string nodeid and
            # do lookup on worker side
            item = [t.nodeid for t in test] if isinstance(test, list) else test.nodeid
            self._test_q.put(item)

        read_results()
        for _ in self._processes:
            self._test_q.put(None)  # signals end

        while any([q is not None for q in self._result_qs]):
            read_results(timeout=1)
        self._test_q.close()
        sys.stdout.write("\r\n")
        self._output_summary()


if __name__ == "__main__":
    args = sys.argv[1:]
    plugins_to_load = []
    config = _prepareconfig(args, plugins_to_load)
    config.hook.pytest_cmdline_main(config=config)



import os
import socket
import sys
import time
from datetime import datetime

import pytest
from _pytest.config import _prepareconfig, ConftestImportFailure
import pytest_cov.embed

if sys.version_info[0] < 3:
    # noinspection PyUnresolvedReferences
    from Queue import Empty
else:
    # noinspection PyUnresolvedReferences
    from queue import Empty

try:
    my_cov = None  # overriden later, after fork
    try:
        original_cleanup = pytest_cov.embed.cleanupgit
    except:
        original_cleanup = None

    def _cleanup(cov=None):
        # pytest_cov will do special things when tests use multiprocessing, however,
        # here we don't want that processing to take place.  (It will otherwise
        # spit out many lines of messaging all of which has no bearing on success of code coverage
        # collecion)
        if cov != my_cov and original_cleanup:
            original_cleanup(cov)


    pytest_cov.embed.cleanup = _cleanup
    pycov_present = True

except Exception:
    pycov_present = False


class WorkerSession:
    """
    Handles reporting of test status and the like
    """

    def __init__(self, index, result_q, test_generator):
        self._result_q = result_q
        self._index = index
        self._test_generator = test_generator

        self._name = "worker-%d" % (index + 1)
        self._reports = []
        self._count = 0
        self._session_start_time = time.time()
        self._buffered_results = []
        self._buffer_size = 10
        self._timestamp = datetime.utcnow()
        self._last_execution_time = time.time()

    def _put(self, kind, data):
        """
        Append test result data to queue, flushing buffered results to queue at watermark level for efficiency

        :param kind: kind of data
        :param data: test result data to publih to queue
        """
        self._buffered_results.append((kind, data))
        if len(self._buffered_results) >= self._buffer_size or (datetime.utcnow() - self._timestamp).total_seconds() > 1:
            self._flush()

    def _flush(self):
        """
        fluh buffered results out to the queue.
        """
        if self._buffered_results:
            self._result_q.put(self._buffered_results)
            self._buffered_results = []
            self._timestamp = datetime.utcnow()

    def test_loop(self, session):
        """
        This is where the action takes place.  We override the usual implementation since
        that doesn't support a dynamic generation of tests (whose source is the test Queue
        that it draws from to pick out the next test

        :param session:  Where the tests generator is kept
        """
        if pycov_present:
            global my_cov
            try:
                my_cov = pytest_cov.embed.active_cov
            except:
                my_cov = None

        if session.testsfailed and not session.config.option.continue_on_collection_errors:
            raise session.Interrupted("%d errors during collection" % session.testsfailed)

        if session.config.option.collectonly:
            return  # should never really get here, but for consistency

        for preitem in session.items:
            # test item comes through as a unique string nodeid of the test
            # We use the pytest-collected mapping we squirrelled away to look up the
            # actual _pytest.pythone.Function test callable
            for item in preitem if isinstance(preitem, list) else [preitem]:
                item = session._named_items[item]
                item.config.hook.pytest_runtest_protocol(item=item, nextitem=None)
                # very much like that in _pytest.main:
                try:
                    if session.shouldfail:
                        raise session.Failed(session.shouldfail)
                except AttributeError:
                    pass  # some version of pytest do not have this attribute
                if session.shouldstop:
                    raise session.Interrupted(session.shouldstop)
                # count tests that have been run
                self._count += 1
                self._last_execution_time = time.time()


    @pytest.mark.tryfirst
    def pytest_collection_finish(self, session):
        """
        Invoked once pytest has collected all information about which tests to run.
        Those items are squirrelled away in a different attributes, and a generator
        is put in its place to draw from.

        :param session: the pytest session

        :return: the generator of tests to run
        """
        session._named_items = {item.nodeid: item for item in session.items}
        session.items = self._test_generator
        return session.items

    def pytest_internalerror(self, excrepr):
        self._put('error_message', excrepr)

    @pytest.mark.try_first
    def pytest_runtest_logreport(self, report):
        """
        report only status of calls (not setup or teardown), unless there was an error in those
        stash report for later end-game parsing of failures

        :param report: report to draw info from
        """
        self._put('test_status', report)

    def pytest_sessionfinish(self, exitstatus):
        """
        output failure information and final exit status back to coordinator

        :param exitstatus: exit status of the test suite run
        """
        self._put('exit', (self._index, self._count, exitstatus, self._last_execution_time - self._session_start_time))
        self._flush()
        self._result_q.close()


def main(index, test_q, result_q, num_processes):
    """
    main worker function, launched as a multiprocessing Process

    :param index: index assigned to the worker Process
    :param test_q: to draw test names for execution
    :param result_q: to post status and results
    """
    args = sys.argv[1:]
    try:
        def generator():
            try:
                test = test_q.get()
                while test:
                    yield test
                    test = test_q.get()
            except Empty:
                raise StopIteration

        try:
            # use pytest's usual method to prepare configuration
            # we have to re-generate the config in this worker Process, as we cannot pickle the more complex
            # _pytest.python.Function objects across the Queue interface.  This is a slight bit
            # of awkwardness and some overhead per process, but this overhead pales in significance
            # to xdist's use of rsync
            config = _prepareconfig(args, plugins=[])
        except ConftestImportFailure as e:
            result_q.put(('exception', e))
        # unregister terminal (don't want to output to stdout from worker)
        # as well as xdist (don't want to invoke any plugin hooks from another distribute testing plugin if present)
        config.pluginmanager.unregister(name="terminal")
        # register our listener, and configure to let pycov-test knoew that we are a slave (aka worker) thread
        worker = WorkerSession(index, result_q, generator())
        config.pluginmanager.register(worker, "mproc_worker")
        config.option.mproc_worker = worker
        # setup slave info;  this is important to have something defined for these fields
        # so that pycov can pickup on the distributed nature of the test run,
        # and that this a worker(slave) process
        workerinput = {'slaveid': "worker-%d" % index,
                       'workerid': "worker-%d" % index,
                       "workercount": num_processes,
                       "slavecount": num_processes,
                       'cov_master_host': socket.gethostname(),
                       'cov_slave_output': os.path.join(os.getcwd(), "worker-%d" % index),
                       'cov_master_topdir': os.getcwd()
                       }
        config.slaveinput = workerinput
        config.slaveoutput = workerinput
        try:
            # and away we go....
            config.hook.pytest_cmdline_main(config=config)
        finally:
            config._ensure_unconfigure()
    except Exception as e:
        result_q.put(('exception', e))
        result_q.put(('exit', (index, -1, -1)))




import inspect


def group(name):
    """
    Decorator for grouping tests or a class of tests
    :param name: unique name for group of tests to be serialized under execution
    """
    def decorator_group(object):
        if inspect.isclass(object):
            for method in [getattr(object, m) for m in dir(object) if inspect.isfunction(getattr(object, m)) and m.startswith('test')]:
                method._pytest_group = name
        elif inspect.isfunction(object):
            object._pytest_group = name
        else:
            raise Exception("group decorator can only decorate class or function object")
        return object
    return decorator_group



import setuptools

VERSION="1.4.4"
setuptools.setup(
    name='pytest_mproc',
    author='John Rusnak',
    author_email='jrusnak@linkedin.com',
    version=VERSION,
    package_dir={'': 'src'},
    packages=setuptools.find_packages('src'),
    entry_points={
       "pytest11": ["name_of_plugin = pytest_mproc.plugin"],
    },
    classifiers=["Framework :: Pytest",
                 "Development Status :: 4 - Beta",
                 "License :: OSI Approved :: BSD License"],
    license='BSD 2-CLAUSE',
    keywrds='pytest distributed multiprocessing',
    url='https://github.com/jrusnakli/pytest_mpdist',
    download_url="https://github.com/jrusnakli/pytest_mpdist/dist/%s" % VERSION,
)



import os
import pytest_mproc
import sys
# to pick up dummy_src (ensure in path):
sys.path.append(os.path.basename(__file__))
sys.path.append(os.path.join(os.path.basename(__file__), "..", "testcode"))

import pytest

from testcode.dummy_src.something import to_be_run_under_test
from testcode.testsomething import Something
from pytest_mproc.utils import global_finalizer, global_initializer, global_session_context

@pytest.fixture()
def some_fixture():
    return 1

@pytest_mproc.group("group")
def test_group_m2():
    Something().group_m1()
    assert Something.proc_id == os.getpid()


@pytest_mproc.group("class_group")
class TestGrouped:

    proc_id = None

    def test1(self):
        TestGrouped.proc_id = os.getpid()

    def test2(self):
        assert TestGrouped.proc_id == os.getpid()


def test_some_alg1():
    Something().some_alg1()


@pytest.mark.parametrize('data', ['a%s' % i for i in range(1000)])
def test_some_alg2(data, some_fixture):
    Something().some_alg2(data)


def test_some_alg3():
    Something().some_alg3()


@global_initializer("some_fixture")
def initializer():
    print("\n###### INITIALIZE ############\n")


@global_finalizer("some_fixture")
def finalizer():
    print("\n###### FINALIZE ############\n")


@global_session_context()
def sess_context():
    print("\n>>> IN GLOBAL SESSION CONTEXT\n")
    yield
    print("\n>>> OUT OF GLOBAL SESSION CONTEXT\n")


@global_session_context("some_fixture")
class CM:

    def __enter__(self):
        print("\n>>> IN CLASS GLOBAL SESSION CONTEXT\n")

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("\b>>> OUT OF CLASS GLOBAL SESSION CONTEXT\n")
